int PDC_clearclipboard( void)
{
   return( 0);
}

int PDC_freeclipboard( char *contents)
{
   return( 0);
}

int PDC_getclipboard( char **contents, long *length)
{
   return( 0);
}

int PDC_setclipboard( const char *contents, long length)
{
   return( 0);
}

